/*!
 * © 2014, 2022 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals Character: true, alert, Rules, Genre, options, saveAs */
Character = {};
var character;

Character.allstats = [ "str", "dex", "agi", "con", "app", "int", "wit", "will" ];
Character.longstats = [ "Strength", "Dexterity", "Agility", "Constitution",
        "Appearance", "Intellect", "Wit", "Will" ];

Character.returnFalse = function(f) {
    return function() {
        try {
            f();
        } catch (e) {
            if (e.stack) {
                alert(e + "\n\nStack Trace\n-----------------\n" + e.stack);
            } else {
                alert(e);
            }
        }
        return false;
    };
};

Character.incrementVersion = function(cid) {
    character[cid].characterVersion += 1;
    character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
    Character.populateJSON(cid);
};

Character.updateDays = function(cid) {
    character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime, character[cid].birthmonth);
    character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
};

Character.templates = {
    interest : {
        skill : "Numeracy",
        ip : 1,
        teacher : false
    }
};

Character.setIdText = function(cid, id, text, tooltip) {
    var element = document.getElementById("c"+cid+"-"+id);
    var newText = document.createTextNode(text);
    if (tooltip) {
        element.title = tooltip;
    }
    if (element.hasChildNodes()) {
        element.replaceChild(newText, element.firstChild);
    } else {
        element.appendChild(newText);
    }
};
Character.setIdClass = function(cid, id, cls) {
    var element = document.getElementById("c"+cid+"-"+id);
    element.className = cls;
};
Character.setTextArea = function(cid, field, text) {
    var element = document.getElementById("c"+cid+"-"+field);
    element.value = text;
    element.defaultValue = text;
};

Character.setOption = function(options, choice) {
    var found = false;
    for (var j = 0; j < options.length; j++) {
        if (choice.toString() === options[j].text) {
            options.selectedIndex = j;
            found = true;
        }
    }
    if (!found) {
        var newOption = new Option(choice.toString(), choice.toString(), true,
                true);
        options.add(newOption);
    }
};

Character.setChecked = function(button, choice) {
    button.checked = choice;
    button.defaultChecked = choice;
};

Character.deleteElement = function(cid, idOrElement) {
    var element = ((typeof idOrElement === "string") ? document.getElementById("c"+cid+"-"+idOrElement) : idOrElement);
    if (element) {
        element.parentNode.removeChild(element);
    }
};

Character.durationLength = function(dur) {
    if (isNaN(dur)) {
        throw new Error("Duration is not a number");
    } else if (dur === 0) {
        return "0 months";
    } else if (dur === 1) {
        return "1 month";
    } else if (dur < 12) {
        return dur + " months";
    }
    var years = Math.floor(dur / 12);
    dur = dur % 12;

    if (years === 1) {
        if (dur === 0) {
            return "1 year";
        }
        return "1 year, " + Character.durationLength(dur);
    } else {
        if (dur === 0) {
            return years + " years";
        }
        return years + " years, " + Character.durationLength(dur);
    }
};

Character.getFluencyPoints = function(cid,p) {
    var endFluency = (character[cid].periods[p].endmonth < 179) ? character[cid].periods[p].endmonth
            : 179;
    var fluencyMonths = endFluency - character[cid].periods[p].startmonth + 1;
    if (fluencyMonths < 0) {
        return 0;
    }
    var points = (character[cid].periods[p].startmonth === 60) ? 1000 : 0;
    points += fluencyMonths * 8;
    for (var anniversary = 72; anniversary <= 180; anniversary += 12) {
        if ((character[cid].periods[p].startmonth <= anniversary) && (character[cid].periods[p].endmonth >= anniversary)) {
            points += 4;
        }
    }
    return points;
};

Character.updateFluencyPoints = function(cid,p) {
    var points = Character.getFluencyPoints(cid,p);
    if (points > 0) {
        Character.setIdText(cid,"period"+p+"-fluencypoints",points+" points");
    }
};

Character.deleteAllChildren = function(node) {
    while (node.firstChild) {
        node.removeChild(node.firstChild);
    }
};

Character.addInterestRow = function(cid, p,table,index) {
    var template = document.getElementById("c"+cid+"-period@@-interest0");
    var row = template.cloneNode(true);
    var next = index;
    Character.updateNode(row,p);
    Character.updateForm(row,next);
    row.id = "c"+cid+"-period"+p+"-interest"+next;
    table.appendChild(row);
    Character.addSkillRow(cid,p,next);
    Character.populateJSON(cid);
};

Character.addInterest = function(cid,skillname,ip) {
    return Character.returnFalse(function() {
        // Work out which tab is active
        var tab;
        var tabbersheet = document.getElementById("c"+cid+"-sheet");
        if (tabbersheet.tabber) {
            tab = tabbersheet.tabber.getActiveTab();
        }
        if (tab) {
            ip = ip || 1;
            var idbits = tab.id.split("-"); // Note we assume an ordering of cid, period, interest as set up by addInterestRow
            if (idbits[1].slice(0,6) === "period") {
                var p = Number(idbits[1].slice(6));
                var table = document.getElementById("c"+cid+"-period"+p+"-details").firstChild.nextSibling;
                if (character[cid].periods[p].iptotal+ip <= character[cid].interestPoints) {
                    Character.addInterestKnowingPeriod(cid,p,table,skillname,ip);
                    Character.incrementVersion(cid);
                    Character.updateStatus(cid,p);
                }
            }
        }
    });
};

Character.updateTabs = function(cid) {
    var tabbersheet = document.getElementById("c"+cid+"-sheet");
    if (tabbersheet.tabber) {
        tabbersheet.tabber.reinit(tabbersheet);
    }
};

Character.updateArmour = function(cid,armour) {
    return Character.returnFalse(function() {
        character[cid].armour = armour;
        Character.incrementVersion(cid);
        Character.populateJSON(cid);
    });
};

Character.updateShield = function(cid,shield) {
    return Character.returnFalse(function() {
        character[cid].shield = shield;
        if (character[cid].shield.locations === 0) {
            character[cid].shieldLocations = [];
        } else if (character[cid].shield.locations === 1) {
            character[cid].shieldLocations = [ "shieldArm", "chest"];
        } else if (character[cid].shield.locations === 2) {
            character[cid].shieldLocations = [ "shieldArm", "chest","head"];
        } else if (character[cid].shield.locations === 3) {
            character[cid].shieldLocations = [ "shieldArm", "chest","head","abdomen"];
        }
        Character.populateShieldLocations(cid);
        Character.incrementVersion(cid);
        Character.populateJSON(cid);
    });
};

Character.updateShieldLocations = function(cid) {
    return Character.returnFalse(function() {
        var locs;
        locs = ["head", "chest", "abdomen"];
        character[cid].shieldLocations = ["shieldArm"];
        for(var i=0; i<locs.length; i++) {
            if (document.getElementById("c"+cid+"-coverage"+locs[i]).checked) {
                character[cid].shieldLocations.push(locs[i]);
            }
        }
        Character.incrementVersion(cid);
    });
};

Character.populateShieldLocations = function(cid) {
    var containsShieldLocation = function(location) {
        return function(x) { return x === location; };
    };
    document.getElementById("c"+cid+"-trcoverageinfo").firstChild.data = character[cid].shield.text;
    var head =  document.getElementById("c"+cid+"-coveragehead");
    var chest = document.getElementById("c"+cid+"-coveragechest");
    var abdomen = document.getElementById("c"+cid+"-coverageabdomen");
    if (character[cid].shield.locations === 0) {
        head.type = "checkbox";
        chest.type = "checkbox";
        abdomen.type = "checkbox";
        head.checked = false;
        chest.checked = false;
        abdomen.checked = false;
        head.disabled = true;
        chest.disabled = true;
        abdomen.disabled = true;
    } else if (character[cid].shield.locations === 1) {
        head.type = "radio";
        chest.type = "radio";
        abdomen.type = "radio";
        head.checked = character[cid].shieldLocations.some(containsShieldLocation("head"));
        chest.checked = character[cid].shieldLocations.some(containsShieldLocation("chest"));
        abdomen.checked = character[cid].shieldLocations.some(containsShieldLocation("abdomen"));
        head.disabled = false;
        chest.disabled = false;
        abdomen.disabled = false;
        head.onchange = Character.updateShieldLocations(cid);
        chest.onchange = Character.updateShieldLocations(cid);
        head.onchange = Character.updateShieldLocations(cid);
    } else if (character[cid].shield.locations === 2) {
        head.type = "radio";
        chest.type = "checkbox";
        abdomen.type = "radio";
        head.checked = character[cid].shieldLocations.some(containsShieldLocation("head"));
        chest.checked = true;
        abdomen.checked = character[cid].shieldLocations.some(containsShieldLocation("abdomen"));
        head.disabled = false;
        chest.disabled = true;
        abdomen.disabled = false;
        head.onchange = Character.updateShieldLocations(cid);
        abdomen.onchange = Character.updateShieldLocations(cid);
    } else if (character[cid].shield.locations === 3) {
        head.type = "checkbox";
        chest.type = "checkbox";
        abdomen.type = "checkbox";
        head.checked = true;
        chest.checked = true;
        abdomen.checked = true;
        head.disabled = true;
        chest.disabled = true;
        abdomen.disabled = true;
    }
};

Character.updateFavouredWeapons = function(cid) {
    return Character.returnFalse(function() {
        character[cid].favouredWeapons = [];
        for(var i=0; i<Rules.weapons.length; i++) {
            var checkbox = document.getElementById("c"+cid+"-weapon-"+Rules.weapons[i].name);
            if (checkbox && checkbox.checked) {
                character[cid].favouredWeapons.push(Rules.weapons[i]);
            }
        }
        Character.incrementVersion(cid);
        Character.populateJSON(cid);
    });
};

Character.populateEquipment = function(cid) {
    var containsWeapon = function(weapon) { return function(x) { return x.name === weapon.name; }; };
    var containsGenre = function(x) { return x === character[cid].genre.toString(); };
    var radioequip = function(radioname,array,setting,update) {
        var b = 0;
        var button,span;
        for(var i=0; i<array.length; i++) {
            if (array[i].genre.some(containsGenre)) {
                span = document.getElementById("c"+cid+"-td"+radioname+b);
                button = document.getElementById("c"+cid+"-"+radioname+b);
                span.className = "";
                button.nextSibling.data = array[i].name;
                button.checked = (array[i].name === setting.name);
                button.onchange = update(cid,array[i]);
                b++;
            }
        }
        // Clear out the other buttons, if any
        for(;;b++) {
            span = document.getElementById("c"+cid+"-td"+radioname+b);
            if (!span)  { break; }
            button = document.getElementById("c"+cid+"-"+radioname+b);
            span.className = "hidden";
            if (button.checked) {
                button.checked = false;
                document.getElementById("c"+cid+"-"+radioname+"0").checked = true;
            }
        }

    };

    radioequip("armour",Rules.armour,character[cid].armour,Character.updateArmour);
    radioequip("shield",Rules.shields,character[cid].shield,Character.updateShield);

    // shield locations
    Character.populateShieldLocations(cid);

    // Weapons are a set of checkboxs, (6 columns)
    var columns = 6;
    var weapons = document.getElementById("c"+cid+"-weapons");
    var newweapons = document.createDocumentFragment();
    var newrow = document.createElement("tr");
    newrow.id = "c"+cid+"-weapons";
    var column = 0;
    for(var i=0; i<Rules.weapons.length; i++) {
        if (Rules.weapons[i].genre.some(containsGenre)) {
            var newcheckbox = document.createElement("input");
            newcheckbox.type = "checkbox";
            newcheckbox.id = "c"+cid+"-weapon-"+Rules.weapons[i].name;
            newcheckbox.checked = character[cid].favouredWeapons.some(containsWeapon(Rules.weapons[i]));
            newcheckbox.onchange = Character.updateFavouredWeapons(cid);
            var newcell = document.createElement("td");
            newcell.appendChild(newcheckbox);
            newcell.appendChild(document.createTextNode(Rules.weapons[i].name));
            newrow.appendChild(newcell);
            column++;
            if (column === columns) {
                newweapons.appendChild(newrow);
                newrow = document.createElement("tr");
                column = 0;
            }
        }
    }
    if (column > 0) {
        newweapons.appendChild(newrow);
    }
    while(weapons.nextSibling) {
        weapons.parentNode.removeChild(weapons.nextSibling);
    }
    weapons.parentNode.replaceChild(newweapons,weapons);
};

Character.populatePeriodMonths = function(cid,p) {
    Character.setIdText(cid,"period"+p+"-enddatelength"," A period of "+Character.durationLength(character[cid].periods[p].endmonth - character[cid].periods[p].startmonth  + 1));
    if (character[cid].periods[p].startmonth === character[cid].periods[p].endmonth) {
        Character.setIdText(cid,"period"+p+"-daterange","The month of "+character[cid].periods[p].startmonthtext);
    } else if (character[cid].periods[p].startmonth < character[cid].periods[p].endmonth) {
        Character.setIdText(cid,"period"+p+"-daterange",character[cid].periods[p].startmonthtext+" to "+character[cid].periods[p].endmonthtext);
    } else {
        Character.setIdText(cid,"period"+p+"-daterange","Future interests");
    }
};

Character.populateSkills = function(cid) {
    var i,j;
    var endingMonthText = character[cid].genre.dateMonthAsText(character[cid].periods[character[cid].periods.length-1].endmonth+character[cid].birthmonth);

    var points = {};
    var addPoints = function(points,name,dp) {
        if (points[name]) {
            points[name] += dp;
        } else {
            points[name] = dp;
        }
    };
    for(i=0; i<character[cid].periods.length; i++) {
        // Childhood fluency
        if (Character.getFluencyPoints(cid,i) > 0) {
            addPoints(points,"Fluency("+character[cid].periods[i].childhoodLanguage+")",Character.getFluencyPoints(cid,i));
        }
        var duration = character[cid].periods[i].endmonth - character[cid].periods[i].startmonth + 1;
        for(j=0; j<character[cid].periods[i].interests.length; j++) {
            var dp = character[cid].periods[i].interests[j].ip * duration;
            if (character[cid].periods[i].interests[j].teacher) {
                dp += (Math.round(character[cid].periods[i].interests[j].ip / 3 - 0.3)) * duration;
            }
            addPoints(points,character[cid].periods[i].interests[j].skill,dp);
        }
    }
    // convert to array and sort
    var arr = [];
    for(var s in points) {
        if (points.hasOwnProperty(s)) {
            arr.push({skill:s, dp: points[s]});
        }
    }
    // Compute the skill levels
    for(i=0; i<arr.length; i++) {
        arr[i].level = Rules.skilllevel(arr[i].skill,character[cid],arr[i].dp);
    }


    arr.sort(function(x,y) {
        return y.level - x.level;
    });
    var DOM_table = document.createElement("table");
    var DOM_tbody = document.createElement("tbody");
    var DOM_tr = document.createElement("tr");
    var DOM_th = document.createElement("th");
    DOM_th.colspan = 3;
    var DOM_p = document.createElement("p");
    DOM_p.appendChild(document.createTextNode("As of the month of "+endingMonthText));
    DOM_th.appendChild(DOM_p);
    DOM_tr.appendChild(DOM_th);
    DOM_tbody.appendChild(DOM_tr);

    DOM_tr = document.createElement("tr");
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Skill Name"));
    DOM_tr.appendChild(DOM_th);
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Development Points"));
    DOM_tr.appendChild(DOM_th);
    DOM_th = document.createElement("th");
    DOM_th.appendChild(document.createTextNode("Level"));
    DOM_tr.appendChild(DOM_th);
    DOM_tbody.appendChild(DOM_tr);

    for(i=0; i<arr.length; i++) {
        DOM_tr = document.createElement("tr");
        var DOM_td = document.createElement("td");
        DOM_td.appendChild(document.createTextNode(arr[i].skill));
        DOM_tr.appendChild(DOM_td);
        DOM_td = document.createElement("td");
        DOM_td.style = "text-align: right";
        DOM_td.appendChild(document.createTextNode(arr[i].dp));
        DOM_tr.appendChild(DOM_td);
        DOM_td = document.createElement("td");
        DOM_td.style = "text-align: right";
        DOM_td.appendChild(document.createTextNode(arr[i].level));
        DOM_tr.appendChild(DOM_td);
        DOM_tbody.appendChild(DOM_tr);
    }
    DOM_table.appendChild(DOM_tbody);

    var skills = document.getElementById("c"+cid+"-skillsheet");
    Character.deleteAllChildren(skills);
    skills.appendChild(DOM_table);

    // And now the JSON needs updating
    character[cid].sheet = { endmonth: character[cid].periods[character[cid].periods.length-1].endmonth,
                        endmonthtext: endingMonthText,
                        skills: arr
                      };
    Character.populateJSON(cid);
};

Character.validateStat = function(name,stat) {
    if (stat === undefined) { throw name+" is blank, character has not been updated."; }
    var v = Number(String(stat).trim());
    if (isNaN(v)) { throw name+" is not a number (it's '"+stat+"'). character has not been updated."; }
    if (v < 0)    { throw name+" is less than zero, character has not been updated."; }
    if (v > 20)   { throw name+" is more than twenty, character has not been updated."; }
    return v;
};

Character.createButton = function(name,onclickfunction) {
    var DOM_button = document.createElement("button");
    var text = name.replace("/","/\u200b").replace("(","\u200b(");
    DOM_button.appendChild(document.createTextNode(text));
    DOM_button.value = name;
    DOM_button.style.width = "100%";
    DOM_button.onclick = onclickfunction;
    return DOM_button;
};

Character.createTextArea = function(cid, id, rows, cols) {
    var DOM_textarea = document.createElement("textarea");
    DOM_textarea.id= "c"+cid+"-"+id;
    DOM_textarea.rows = rows;
    DOM_textarea.cols = cols;
    if (rows === 1) {
        DOM_textarea.className = "onerow";
    }
    DOM_textarea.appendChild(document.createTextNode(""));
    return DOM_textarea;
};

Character.updateStatus = function(cid,p) {
    var status = document.getElementById("c"+cid+"-period" + p + "-status");
    var text;
    var valid = false;
    var DOM_b;
    Character.deleteAllChildren(status);
    if (character[cid].periods[p].endmonth === character[cid].periods[p].startmonth-1) {
        status.setAttribute("class", "warning");
        text = "This period is of zero length and will not add any development points to this character's skills.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].endmonth < character[cid].periods[p].startmonth) {
        status.setAttribute("class", "error");
        text = "This period is of NEGATIVE length and will REMOVE development points from this character's skills.";
        DOM_b = document.createElement("b");
        DOM_b.appendChild(document.createTextNode(text));
        status.appendChild(DOM_b);
    } else if (p === character[cid].periods[p].length-1 && character[cid].periods[p].endMonth !== character[cid].gamemonth-character[cid].birthmonth) {
        status.setAttribute("class", "warning");
        text = "The end month of this last period (" + character[cid].periods[p].endMonthText + ") does not match the game month (" + character[cid].gamemonthtext + "). When the game month is updated, this will reset this end month to the game month.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal === 0) {
        status.setAttribute("class", "warning");
        text = "None of the " + character[cid].interestPoints + " interest points available to this character have been used.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal < character[cid].interestPoints) {
        status.setAttribute("class", "warning");
        text = "Only " + character[cid].periods[p].iptotal + " of the " + character[cid].interestPoints + " interest points available to this character have been used.";
        status.appendChild(document.createTextNode(text));
    } else if (character[cid].periods[p].iptotal === character[cid].interestPoints) {
        status.setAttribute("class", "");
        text = "All the " + character[cid].interestPoints + " interest points available to this character have been used.";
        valid = true;
        status.appendChild(document.createTextNode(text));
    } else {
        status.setAttribute("class", "error");
        text = "The " + character[cid].periods[p].iptotal + " interest points used by this character are more than the " + character[cid].interestPoints + " available to this character[cid].";
        DOM_b = document.createElement("b");
        DOM_b.appendChild(document.createTextNode(text));
        status.appendChild(DOM_b);
    }
    character[cid].periods[p].statustext = text;
    character[cid].periods[p].valid = valid;
};

Character.populateAll = function(cid) {
    Character.populateStats(cid);
    Character.populateEquipment(cid);
    Character.populatePeriods(cid);
    Character.populateSkills(cid);
};


Character.initPage = function(cid) {
    character = character || [];
    // Initialise the character, either with data from an embedded JSON or a
    // completely new one
    if (document.getElementById("c"+cid+"-character-data") && document.getElementById("c"+cid+"-character-data").text && document.getElementById("c"+cid+"-character-data").text.length > 10) {
        Character.updateWithNewJSON(cid, document.getElementById("c"+cid+"-character-data").text, false);
    } else {
        character[cid] = Rules.newCharacter();
        character[cid].schemaVersion = 13;
        character[cid].characterVersion = 0;
        character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
        character[cid].rulesVersion = "2014a";
        character[cid].genre = new Genre(options.default_genre);
        character[cid].name = "Nobody Noname";
        character[cid].gender = '';
        character[cid].playerName = document.getElementById("c"+cid+"-playernametextarea").textContent;
        character[cid].description = "An outstanding character whose deeds are described herein and whose heroic future is still to come.";
        character[cid].statslock = false;
        character[cid].statsswapable = true;
        character[cid].morale = Rules.startingMorale(character[cid]);
        character[cid].woundPoints = Rules.woundPoints(character[cid]);
        character[cid].interestPoints = Rules.interestPoints(character[cid]);
        character[cid].birthtime = Date.now() - 32 * 365.25 * 24 * 60 * 60 * 1000;
        character[cid].birthmonth = character[cid].genre.asMonth(character[cid].birthtime);
        character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
        character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime, character[cid].birthmonth);
        character[cid].gametime = options.gametime[character[cid].genre.name];
        character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
        character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime, character[cid].gamemonth);
        character[cid].periods = [ {
            title : "Childhood",
            description : "The early years discovering the world and your place in it",
            startmonth : 60,
            startmonthtext : character[cid].genre.dateMonthAsText(character[cid].birthmonth + 60),
            endmonth : character[cid].gamemonth - character[cid].birthmonth,
            endmonthtext : character[cid].gamemonthtext,
            childhoodLanguage : character[cid].genre.getDefaultLanguage(),
            interests : [],
            iptotal : 0,
            statustext : "None of the " + character[cid].interestPoints + " interest points available to this character have been used.",
            valid : false
        } ];
        character[cid].newSkills = [];
        character[cid].armour = Rules.armour[0];
        character[cid].shield = Rules.shields[0];
        character[cid].shieldLocations = [];
        character[cid].favouredWeapons = [];
        character[cid].syncWithGlobalDate = true;
    }
    // Set up all the non-Period buttons
    Character.setupButtons(cid);

    document.getElementById("c"+cid+"-JSONsavefile").onclick = Character
            .returnFalse(function() {
                Character.incrementVersion(cid);
                var blob = new Blob(
                        [ JSON.stringify(character[cid], undefined, 2) ], {
                            type : "text/plain;charset=utf-8"
                        });
                saveAs(blob, character[cid].name + ".txt");
            });

    document.getElementById("c"+cid+"-savePDF").onclick = Character
            .returnFalse(function() {
                var blob = Character.makePDF(cid).output('blob');
                saveAs(blob, character[cid].name + ".pdf");
            });
};

Character.updateWithNewJSON = function(cid,json,populate) {
    try {
        var i;
        var newCharacter = JSON.parse(json);
        for(i=0; i<Character.allstats.length; i++) {
            newCharacter[Character.allstats[i]] = Character.validateStat(Character.allstats[i],newCharacter[Character.allstats[i]]);
        }
        character[cid] = newCharacter;

        if (!character[cid].genre) {
            // Need to guess! Base off of the childhood language in the first period
            if (character[cid].periods[0].childhoodLanguage === "Athionic") {
                character[cid].genre = new Genre("Linrodeth");
            } else if (character[cid].periods[0].childhoodLanguage === "Latin") {
                character[cid].genre = new Genre("Luna Romana");
            } else if (character[cid].periods[0].childhoodLanguage === "English") {
                character[cid].genre = new Genre("Contemporary");
            } else if (character[cid].periods[0].childhoodLanguage === "Tsolyáni") {
                character[cid].genre = new Genre("Tekumel");
            } else if (character[cid].periods[0].childhoodLanguage === "Anglic") {
                character[cid].genre = new Genre("Traveller");
            } else {
                character[cid].genre = new Genre("Linrodeth");
            }
        } else {
            character[cid].genre = new Genre(character[cid].genre);
        }

        character[cid].morale = Rules.startingMorale(character[cid]);
        character[cid].woundPoints = Rules.woundPoints(character[cid]);
        character[cid].interestPoints = Rules.interestPoints(character[cid]);
        if (!character[cid].schemaVersion) {
            character[cid].schemaVersion = 1;
            character[cid].characterVersion = 1e9; // Large offset to make it obvious
            character[cid].lastEditDate = (new Date(Date.now())).toLocaleString();
            character[cid].rulesVersion = "2014a";
            character[cid].description = "An outstanding character whose deeds are described herein.";
            character[cid].gametime = character[cid].birthtime + character[cid].periods[character[cid].periods.length-1].endmonth*30.4375*24*60*60*1000; // Approx
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
            character[cid].gamemonth = character[cid].periods[character[cid].periods.length-1].endmonth + character[cid].birthmonth;
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
            character[cid].periods[character[cid].periods.length-1].endMonth = character[cid].gamemonth;
            character[cid].periods[character[cid].periods.length-1].endMonthText = character[cid].gamemonthtext;
            character[cid].statslock = true;
            Character.updateDays(cid);
        }
        Rules.resetAddedSkills();
        Character.resetOtherSkillsList(cid);

        // Schema update processing
        var name;
        if (character[cid].schemaVersion === 1) {
            character[cid].schemaVersion = 2;
        } else {
            for(i=0; i<character[cid].newSkills.length; i++) {
                for(name in character[cid].newSkills[i]) {
                    if (character[cid].newSkills[i].hasOwnProperty(name)) {
                        Rules.addSkill(name,character[cid].newSkills[i][name]);
                    }
                }
            }
        }
        // Rebuild the newSkills list, (to remove duplicates and to ensure the added skills remain even when building lots of characters from the same instance
        character[cid].newSkills = [];
        for(name in Rules.skilllist["Added skills"]) {
            if (Rules.skilllist["Added skills"].hasOwnProperty(name)) {
                Character.addOtherSkill(cid,name,Rules.skilllist["Added skills"][name]);
            }
        }

        // Continue with the schema updates
        if (character[cid].schemaVersion === 2) {
            character[cid].schemaVersion = 3;
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
        }
        if (character[cid].schemaVersion === 3) {
            character[cid].schemaVersion = 4;
            character[cid].armour = Rules.armour[0];
            character[cid].shield = Rules.shields[0];
            character[cid].shieldLocations = [];
            character[cid].favouredWeapons = [];
        }
        if (character[cid].schemaVersion === 4) {
            character[cid].schemaVersion = 5;
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].birthmonth);
        }
        if (character[cid].schemaVersion === 5) {
            character[cid].schemaVersion = 6;
            character[cid].birthday = character[cid].genre.dayMonthYearAsText(character[cid].birthtime,character[cid].birthmonth);
            character[cid].birthmonth = character[cid].genre.asMonth(character[cid].birthtime);
            character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
            character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        }
        if (character[cid].schemaVersion === 6) {
            character[cid].schemaVersion = 7;
            character[cid].playerName = document.getElementById("c"+cid+"-playernametextarea").textContent;
        }
        if (character[cid].schemaVersion < 9) {
            character[cid].schemaVersion = 9;
            if (character[cid].genre.equals(new Genre('Linrodeth'))) {
                character[cid].birthmonth--;
                character[cid].gamemonth--;
            }
            for(i=0; i<character[cid].periods.length; i++) {
                character[cid].periods[i].startmonthtext = character[cid].genre.dateMonthAsText(character[cid].periods[i].startmonth+character[cid].birthmonth);
                character[cid].periods[i].endmonthtext = character[cid].genre.dateMonthAsText(character[cid].periods[i].endmonth+character[cid].birthmonth);
            }
            character[cid].birthmonthtext = character[cid].genre.dateMonthAsText(character[cid].birthmonth);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
        }
        if (character[cid].schemaVersion === 9) {
            character[cid].schemaVersion = 10;
            character[cid].morale = Rules.startingMorale(character[cid]); // rounding change
        }
        if (character[cid].schemaVersion === 10) {
            character[cid].schemaVersion = 11;
            character[cid].syncWithGlobalDate = (character[cid].gametime > 1200000000000);
        }
        if (character[cid].schemaVersion === 11) {
            character[cid].schemaVersion = 12;
            character[cid].statsswapable = !(character[cid].statslock);
        }
        if (character[cid].schemaVersion === 12) {
            character[cid].schemaVersion = 13;
            character[cid].gender = '';
        }

        // Data resets - resolve skills with
        var ampregexp = new RegExp('&amp;', 'g');
        for(i=0; i<character[cid].periods.length; i++) {
            for(var j=0; j<character[cid].periods[i].interests.length; j++) {
                character[cid].periods[i].interests[j].skill = character[cid].periods[i].interests[j].skill.replace(ampregexp,'&');
            }
        }

        // Global updates
        if (character[cid].syncWithGlobalDate) {
            character[cid].gametime = options.gametime[character[cid].genre.name];
            character[cid].gameday = character[cid].genre.dayMonthYearAsText(character[cid].gametime,character[cid].gamemonth);
            character[cid].gamemonth = character[cid].genre.asMonth(character[cid].gametime);
            character[cid].gamemonthtext = character[cid].genre.dateMonthAsText(character[cid].gamemonth);
            character[cid].periods[character[cid].periods.length-1].endmonth = character[cid].gamemonth - character[cid].birthmonth;
            character[cid].periods[character[cid].periods.length-1].endmonthtext = character[cid].gamemonthtext;
        }
        Character.updateDays(cid); // Get the text for birthday and gameday in sync

        if (populate) {
            Character.populateAll(cid);
        }
    } catch(e) {
        alert(e);
    }
};
