/*!
 * © 2014, 2022 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals Character: true, Rules, jsPDF, character */
Character.makePDF = function(cid) {
    var drawSkillTable = function(pdf, title, skills, top, left, cellheight, cellwidth) {
        if (skills.length > 45) {
            cellheight--; // Cope with Goodgulf :)
        }
        if (skills.length > 50) {
            cellheight -= 0.5; // And again
        }

        var skillnamewidth = 81;
        var skilllevelwidth = 60;
        var levelsplit = 112;
        var vgutter = ((skills.length > 50) ? 2 : 3);
        var hgutter = 2;

        var mainleft = left + skillnamewidth + skilllevelwidth;
        var right = mainleft + Rules.adjectives.length * cellwidth;
        var bottom = top + (skills.length + 2) * cellheight;

        pdf.setLineWidth(1);
        pdf.line(mainleft, top, right, top);
        pdf.line(mainleft, top + cellheight, right, top + cellheight);
        pdf.setLineWidth(0.5);
        for (var x = 0; x < Rules.adjectives.length; x++) {
            pdf.line(mainleft + x * cellwidth, top, mainleft + x * cellwidth,
                    bottom);
            pdf.setFontSize(Rules.adjectives[x].length < 12 ? 6 : 4.5); // Special case for "Embarrasing"
            pdf.textInBox(Rules.adjectives[x], {
                align : "center"
            }, mainleft + x * cellwidth + hgutter, top + cellheight - vgutter,
                    mainleft + (x + 1) * cellwidth, top);
        }

        pdf.setLineWidth(1);
        pdf.line(left, top + cellheight, right, top + cellheight);
        pdf.line(left, top + 2 * cellheight, right, top + 2 * cellheight);
        pdf.line(left, bottom, right, bottom);
        pdf.line(left, top + cellheight, left, bottom);
        pdf.line(left + skillnamewidth, top + cellheight,
                left + skillnamewidth, bottom);
        pdf.line(mainleft, top, mainleft, bottom);
        pdf.line(right, top, right, bottom);
        pdf.setFontStyle("bold");
        pdf.setFontSize(7.5);
        pdf.text(left + hgutter, top + cellheight * 2 - 3, title);
        pdf.text(left + skillnamewidth + hgutter, top + cellheight * 2 - 3, "Level");
        for (x = 0; x < Rules.adjectives.length; x++) {
            pdf.textInBox(x.toString(), {
                align : "center"
            }, mainleft + x * cellwidth + hgutter, top + 2 * cellheight - vgutter,
                    mainleft + (x + 1) * cellwidth, top + cellheight);
        }

        pdf.setLineWidth(0.5);
        pdf.setFontStyle("normal");
        var y = top + 3 * cellheight;
        for (var s = 0; s < skills.length; s++, y += cellheight) {
            if (s !== skills.length - 1) {
                pdf.line(left, y, right, y);
            }
            pdf.text(left + hgutter, y - 3, skills[s].skill);
            pdf.textInBox(skills[s].level.toString(), {
                align : "right"
            }, left + skillnamewidth, y - vgutter, levelsplit - hgutter, y - cellheight + 1);
            pdf.textInBox(Rules.adj(skills[s].level), {
                align : "left"
            }, levelsplit + hgutter, y - vgutter, mainleft, y - cellheight + 1);
            var perf = Rules.performance(skills[s].level);
            for (x = 0; x < Rules.adjectives.length; x++) {
                pdf.textInBox(perf[x], {
                    align : "center"
                }, mainleft + x * cellwidth, y - vgutter, mainleft + (x + 1) * cellwidth, y - cellheight + 1);
            }
        }
    };

    var drawTable = function(pdf, content, widths, aligns, left, top, cellheight, borderweights) {
        var totalwidth = widths.reduce(function(x, y) {
            return x + y;
        }, 0);
        var hgutter = 2;
        pdf.setLineWidth(1);
        pdf.rect(left, top, totalwidth, content.length * cellheight);
        var offset = left + widths[0];
        var cell;
        for (cell = 1; cell < widths.length; cell++) {
            if (borderweights && borderweights[cell]) {
                pdf.setLineWidth(borderweights[cell]);
            }
            pdf.line(offset, top, offset, top + content.length * cellheight);
            offset += widths[cell];
        }
        pdf.setFontStyle("bold"); // Make top line bold
        pdf.setLineWidth(1);
        var y = top + cellheight;
        for (var line = 0; line < content.length; line++, y += cellheight) {
            if (line !== content.length - 1) {
                pdf.line(left, y, left + totalwidth, y);
            }
            offset = left;
            for (cell = 0; cell < aligns.length; cell++) {
                pdf.textInBox(content[line][cell], {
                    align : aligns[cell]
                }, offset + hgutter, y - 3, offset + widths[cell], y - cellheight);
                offset += widths[cell];
            }
            if (line === 0) {
                pdf.setFontStyle("normal");
                pdf.setLineWidth(0.5);
            }
        }
    };

    var pdf = new jsPDF("landscape", "pt", "a4");
    pdf.setProperties({
        title : character[cid].name,
        keywords : "generated, linrodeth, character"
    });

    var namebottom = 30;
    var nameleft = 0;
    var nameright = 842;
    var nametop = 12;
    var left = 16;
    var midright = 400;
    var midleft = 442;
    var right = 842;
    var cellwidth = 32;
    var cellheight = 10;
    var chartext = "Player: " + character[cid].playerName + ", Rules Version: " + character[cid].rulesVersion + ", Plugin Version: 3.214.0, Character version: " + character[cid].characterVersion + ", Game Date:" + character[cid].gameday + ", (last edit time " + character[cid].lastEditDate + ", PDF created " + (new Date()).toLocaleString() + ")";
    var widths;
    var aligns;
    var title;
    var content;

    // Page 1 - back cover skills, levels and current interest
    var skillsummmarycellheight = 12;
    if (character[cid].sheet.skills.length > 32) { skillsummmarycellheight = 11;  }
    if (character[cid].sheet.skills.length > 40) { skillsummmarycellheight = 10;  }
    if (character[cid].sheet.skills.length > 45) { skillsummmarycellheight = 9;   }
    if (character[cid].sheet.skills.length > 50) { skillsummmarycellheight = 8.5; }

    pdf.setFontSize(18);
    pdf.textInBox(character[cid].name, {
        align : "center"
    }, left, namebottom, midright, nametop);
    pdf.setFontSize(skillsummmarycellheight-1);
    content = [];
    widths = [198,100,90];
    aligns = ["left","left", "center"];
    title = ["Skill", "Level", "Current Interest"];
    content.push(title);
    for (var s = 0; s < character[cid].sheet.skills.length; s++) {
        var currentip = "";
        for(var cs = 0; cs < character[cid].periods[character[cid].periods.length-1].interests.length; cs++) {
            if (character[cid].periods[character[cid].periods.length-1].interests[cs].skill === character[cid].sheet.skills[s].skill) {
                currentip = String(character[cid].periods[character[cid].periods.length-1].interests[cs].ip);
            }
        }
        var level = " " + character[cid].sheet.skills[s].level + " (" + Rules.adj(character[cid].sheet.skills[s].level)+")";
        content.push([character[cid].sheet.skills[s].skill, level, currentip]);
    }
    drawTable(pdf,content,widths,aligns,left,namebottom+8,skillsummmarycellheight);

    // front cover name, dates, stats, body, summary
    pdf.setFontSize(18);
    pdf.textInBox(character[cid].name, {
        align : "center"
    }, midleft, namebottom, right, nametop);

    pdf.setFontSize(12);
    content = [];
    widths = [120,100];
    aligns = ["left","left"];
    title = ["Attribute","Level"];
    content.push(title);
    content.push(["Strength"," "+character[cid].str+" ("+Rules.adj(character[cid].str)+")"]);
    content.push(["Dexterity"," "+character[cid].dex+" ("+Rules.adj(character[cid].dex)+")"]);
    content.push(["Agility"," "+character[cid].agi+" ("+Rules.adj(character[cid].agi)+")"]);
    content.push(["Constitution"," "+character[cid].con+" ("+Rules.adj(character[cid].con)+")"]);
    content.push(["Appearance"," "+character[cid].app+" ("+Rules.adj(character[cid].app)+")"]);
    content.push(["Intellect"," "+character[cid].int+" ("+Rules.adj(character[cid].int)+")"]);
    content.push(["Wit"," "+character[cid].wit+" ("+Rules.adj(character[cid].wit)+")"]);
    content.push(["Will"," "+character[cid].will+" ("+Rules.adj(character[cid].will)+")"]);
    content.push(["",""]);
    content.push(["Interest Points"," "+character[cid].interestPoints]);
    content.push(["Morale"," "+character[cid].morale]);
    drawTable(pdf,content,widths,aligns,midleft,namebottom+8,14);

    var midmidleft = (midleft + right)/2+50;
    var bodycentre = midmidleft + (right-midmidleft-20)/2;
    var bodytop = namebottom+8;
    pdf.setLineWidth(1);
    pdf.rect(midmidleft,bodytop,right-midmidleft-20,16*14);
    pdf.setLineWidth(0.5);
    pdf.lines([[32,0],[0,36],[42,0],[0,90],[-20,0],[0,-72],[-7,0],[0,144],[-24,0],[0,-72],[-14,0],[0,72],[-24,0],[0,-144],[-7,0],[0,72],[-20,0],[0,-90],[42,0],[0,-36]],bodycentre-16,bodytop+9);
    var locpic = function(name, location) {
        var containsShieldLocation = function(location) {
            return function(x) {
                return x === location;
            };
        };

        var armour = character[cid].armour.name;
        var points = character[cid].armour.value;
        if (character[cid].shieldLocations.some(containsShieldLocation(location))) {
            armour += " + " + character[cid].shield.name;
            points += character[cid].shield.value;
        }
        return character[cid].woundPoints[location]+"/"+points.toString();
    };
    pdf.textInBox(locpic("Head", "head"),{align : "center", valign: "center"},bodycentre-16,bodytop+45,bodycentre+16,bodytop+9);
    pdf.textInBox(locpic("Shield Arm", "shieldArm"),{align : "center", valign: "center"},bodycentre+39,bodytop+135,bodycentre+57,bodytop+45);
    pdf.textInBox(locpic("Sword Arm", "swordArm"),{align : "center", valign: "center"},bodycentre-59,bodytop+135,bodycentre-39,bodytop+45);
    pdf.textInBox(locpic("Chest", "chest"),{align : "center", valign: "center"},bodycentre-32,bodytop+111,bodycentre+32,bodytop+63);
    pdf.textInBox(locpic("Abdomen", "abdomen"),{align : "center", valign: "center"},bodycentre-32,bodytop+135,bodycentre+32,bodytop+111);
    pdf.textInBox(locpic("Right Leg", "rightLeg"),{align : "center", valign: "center"},bodycentre-32,bodytop+207,bodycentre-8,bodytop+135);
    pdf.textInBox(locpic("Left Leg", "leftLeg"),{align : "center", valign: "center"},bodycentre+6,bodytop+207,bodycentre+32,bodytop+135);

    pdf.text("Gender: "+character[cid].gender,midleft,namebottom+200);
    pdf.text("Birthday: "+character[cid].birthday,midleft,namebottom+214);
    pdf.text("Game Date: "+character[cid].gameday,midleft,namebottom+228);
    pdf.setFontSize(8);
    pdf.textInBox("wound points/armour",{align:"center"}, midmidleft,bodytop+16*14-1.5,right-20,bodytop+16*14-9 );

    pdf.setFontSize(12);
    pdf.setFontStyle("bold");
    pdf.text("Description",midleft,namebottom+260);
    pdf.setFontStyle("normal");
    pdf.text(pdf.splitTextToSize(character[cid].description,right-midleft),midleft,namebottom+274);

    pdf.setFontSize(8);
    pdf.text(left, 576, chartext);

    // Page 2 - inside pages - life biography
    pdf.addPage();
    pdf.setFontSize(8);
    pdf.text(left, 576, chartext);

    var proposedptsize = 12;
    var retry, render = false, done = false;
    var pageoffset, lineoffset;

    while (!done) {
        pageoffset = left;
        lineoffset = 28;
        if (render) { pdf.setFontSize(proposedptsize); }
        retry = false;
        for(var p=0; (p<character[cid].periods.length); p++ ) {
            var desclines = pdf.splitTextToSize(character[cid].periods[p].description,midright-left);
            if (lineoffset + proposedptsize*1.15*(desclines.length+2) > 560) {
                // switch page
                pageoffset += (midleft-left);
                lineoffset = 28;
                if (pageoffset > right) {
                    proposedptsize--;
                    retry = true;
                    break;
                }
            }
            if (render) {
                pdf.setFontStyle("bold");
                pdf.text(character[cid].periods[p].title+" ("+character[cid].periods[p].startmonthtext+"-"+character[cid].periods[p].endmonthtext+")",pageoffset,lineoffset);
                pdf.setFontStyle("normal");
                pdf.text(desclines,pageoffset,lineoffset+proposedptsize*1.15);
            }
            lineoffset += proposedptsize*1.15*(desclines.length+2);
        }
        if (render) { done = true; }
        if (!retry) { render = true; }
    }


    // Page 3 - Misc tables
    pdf.addPage();
    pdf.setFontSize(18);
    pdf.textInBox(character[cid].name, {
        align : "center"
    }, nameleft, namebottom, nameright, nametop);

    // Need to Fake up a skills array for the attributes
    var attributesAsSkills = [];
    for (var stat = 0; stat < Character.allstats.length; stat++) {
        attributesAsSkills.push({
            "skill" : Character.longstats[stat],
            "level" : character[cid][Character.allstats[stat]]
        });
    }
    drawSkillTable(pdf, "Attribute", attributesAsSkills, namebottom+16, left, cellheight, cellwidth);

    var i,w,x,weapon;
    content = [];
    widths = [ 81, 30, 30 ];
    aligns = [ "left", "center", "center" ];
    title = [ "Weapon", "Area", "Hits" ];
    for (x = 0; x < Rules.adjectives.length; x++) {
        title.push(x.toString());
        widths.push(cellwidth);
        aligns.push("center");
    }
    content.push(title);
    var hasarea = false;
    var hashits = false;
    for (w = 0; w < character[cid].favouredWeapons.length; w++) {
        weapon = character[cid].favouredWeapons[w];
        if (weapon.ranged) {
            var area = weapon.area || "";
            var hits = weapon.hits || "";
            hasarea = hasarea || weapon.area || false;
            hashits = hashits || weapon.hits || false;
            var ranges = [ weapon.name, area, hits ];
            ranges = ranges.concat(weapon.ranges);
            for (i = 0; i < ranges.length; i++) {
                ranges[i] = (ranges[i] ? ranges[i].toString() : "");
            }
            content.push(ranges);
        }
    }
    pdf.text(left + 3, 151, "Difficulty to hit at range");
    drawTable(pdf, content, widths, aligns, left, 154, cellheight, [undefined, 1, 0.5, 1, 0.5 ]);

    content = [];
    widths = [ 81, 60 ];
    aligns = [ "left", "center" ];
    title = [ "Weapon", "Damage" ];
    for (x = 0; x < Rules.adjectives.length; x++) {
        title.push(x.toString());
        widths.push(cellwidth);
        aligns.push("center");
    }
    content.push(title);
    var line;
    line = [ "Unarmed", "1" ];
    line = line.concat(Rules.damageTable[1]);
    content.push(line);
    for (w = 0; w < character[cid].favouredWeapons.length; w++) {
        weapon = character[cid].favouredWeapons[w];
        line = [ weapon.name, weapon.damage.toString() ];
        line = line.concat(Rules.damageTable[weapon.damage]);
        content.push(line);
    }
    drawTable(pdf, content, widths, aligns, left, 210, cellheight, [undefined, 1, 1, 0.5 ]);

    content = [];
    content.push([ "Situation", "Mod", "Example for cooking" ]);
    content.push([ "Excellent", "+5", "Staffed restaurant kitchen" ]);
    content.push([ "Very good", "+3", "Skilled help" ]);
    content.push([ "Good", "+1", "Normal kitchen, some help" ]);
    content.push([ "Normal", "", "Normal kitchen" ]);
    content.push([ "Poor", "-1", "No spices or herbs available" ]);
    content.push([ "Very Poor", "-3", "Open fire, few ingredients" ]);
    content.push([ "Terrible", "-5", "No fire" ]);
    drawTable(pdf, content, [ 41, 120 - 86, 225 - 120 ], [ "left", "center",
            "left" ], 45, 324, cellheight);

    var modwidths = [ 90, 36 ];

    content = [];
    content.push([ "Melee situation", "Modifier" ]);
    content.push([ "Defending only", "+3" ]);
    content.push([ "At Opponents side", "+1" ]);
    content.push([ "Behind Opponent", "+2" ]);
    content.push([ "Unarmed versus blunt", "-1" ]);
    content.push([ "Unarmed versus blade", "-4" ]);
    content.push([ "Shorter weapon", "-1" ]);
    content.push([ "Each extra opponent", "-1"]);
    drawTable(pdf, content, modwidths, [ "left", "center" ], 252, 324, cellheight);

    content = [];
    content.push([ "Location Hit", "Melee", "Ranged" ]);
    content.push([ "Head", "91-00", "96-00" ]);
    content.push([ "Left Arm", "81-90", "86-95" ]);
    content.push([ "Right Arm", "71-80", "76-85" ]);
    content.push([ "Chest", "45-70", "51-75" ]);
    content.push([ "Abdomen", "33-44", "33-50" ]);
    content.push([ "Left Leg", "17-32", "17-32" ]);
    content.push([ "Right Leg", "01-16", "01-16" ]);
    drawTable(pdf, content, [ 64, 32, 32 ], [ "left", "center", "center" ],
            405, 324, cellheight);

    content = [];
    content.push([ "Ranged situation", "Modifier" ]);
    content.push([ "Aimed Shot (2 rounds)", "+2" ]);
    content.push([ "Walking target", "-2" ]);
    content.push([ "Jogging target", "-3" ]);
    content.push([ "Running target", "-4" ]);
    content.push([ "Dodging target", "-5" ]);
    content.push([ "Light cover", "-1" ]);
    content.push([ "Medium cover", "-3" ]);
    content.push([ "Heavy cover", "-5" ]);
    drawTable(pdf, content, modwidths, [ "left", "center" ], 252, 410,
            cellheight);

    content = [];
    content.push([ "Modifiers for previous round", "Modifier" ]);
    content.push([ "Minor defeat last round", "-1" ]);
    content.push([ "Clear defeat last round", "-2" ]);
    content.push([ "Major defeat last round", "-3" ]);
    content.push([ "Devastating defeat last round", "-4" ]);
    drawTable(pdf, content, [ 148, 32 ], [ "left", "center" ], 45, 414,
            cellheight);

    content = [];
    content.push([ "Gap", "Type of win", "Wound Location" ]);
    content.push([ "1-3", "Minor", "None" ]);
    content.push([ "4-6", "Clear", "Random" ]);
    content.push([ "7-9", "Major", "Move 1 place, +2 damage" ]);
    content.push([ "10+", "Devastating", "Move 2 places, +5 damage" ]);
    drawTable(pdf, content, [ 32, 120 - left - 32, 225 - 120 ], [ "left",
            "left", "left" ], left, 477 + cellheight, cellheight);
    pdf.setLineWidth(1);
    pdf.setFontStyle("bold");
    pdf.line(left, 477 + cellheight, left, 477);
    pdf.line(left, 477, 225, 477);
    pdf.line(225, 477, 225, 477 + cellheight);
    pdf.text(left+2, 477 + cellheight - 3, "Performance Difference"); // +2 for the hgutter that drawTable used

    pdf.setLineWidth(1);
    pdf.rect(569, 324, 8 * cellwidth, 8 * cellheight);
    pdf.line(569, 324 + cellheight, 569 + 2 * cellwidth, 324 + cellheight);
    pdf.line(569 + 2 * cellwidth, 324, 569 + 2 * cellwidth, 324 + cellheight);
    pdf.setLineWidth(0.5);
    for (i = 1; i < 8; i++) {
        pdf.line(569 + i * cellwidth, 324, 569 + i * cellwidth,
                324 + 8 * cellheight);
        pdf.line(569, 324 + i * cellheight, 569 + 8 * cellwidth, 324 + i * cellheight);
    }
    pdf.text(570, 324 + cellheight - 3, "Morale");
    pdf.text(570 + cellwidth, 324 + cellheight - 3, character[cid].morale.toString());

    content = [];
    content.push([ "Area", "Armour", "WP", "Armour" ]);
    var locline = function(name, location) {
        var containsShieldLocation = function(location) {
            return function(x) {
                return x === location;
            };
        };

        var armour = character[cid].armour.name;
        var points = character[cid].armour.value;
        if (character[cid].shieldLocations.some(containsShieldLocation(location))) {
            armour += " + " + character[cid].shield.name;
            points += character[cid].shield.value;
        }
        return [ name, armour, character[cid].woundPoints[location].toString(),
                points.toString() ];
    };
    content.push(locline("Head", "head"));
    content.push(locline("Shield Arm", "shieldArm"));
    content.push(locline("Sword Arm", "swordArm"));
    content.push(locline("Chest", "chest"));
    content.push(locline("Abdomen", "abdomen"));
    content.push(locline("Right Leg", "rightLeg"));
    content.push(locline("Left Leg", "leftLeg"));
    drawTable(pdf, content, [ 64, 100, 32, 32 ], [ "left", "left", "center",
            "center" ], 405, 410, cellheight, [ 1, 1, 1, 0.5, 1, 0.5 ]);
    pdf.setLineWidth(1);
    pdf.rect(633, 420, 6 * cellwidth, 7 * cellheight);
    pdf.setLineWidth(0.5);
    for (i = 1; i < 7; i++) {
        pdf.line(633, 420 + i * cellheight, 633 + 6 * cellwidth, 420 + i * cellheight);
    }
    for (i = 1; i < 6; i++) {
        pdf.line(633 + i * cellwidth, 420, 633 + i * cellwidth, 420 + 7 * cellheight);
    }

    content = [];
    content.push([ "Type of encumbrance", "Modifier" ]);
    content.push([ "Light encumbrance", "" ]);
    content.push([ "Medium encumbrance", "-1" ]);
    content.push([ "Heavy encumbrance", "-3" ]);
    drawTable(pdf, content, modwidths, [ "left", "center" ], 405, 497,
            cellheight);

    content = [];
    content.push([ "Movement", "Speed", "Modifier" ]);
    content.push([ "Walk", ""+Math.floor(5+character[cid].agi/2), "0" ]);
    content.push([ "Jog", ""+Math.floor(10+character[cid].agi), "-2" ]);
    content.push([ "Sprint", ""+Math.floor(10+3*character[cid].agi), "-4" ]);
    drawTable(pdf, content, [ 64, 32, 64 ], [ "left", "center", "center" ], 570, 497,
            cellheight);

    if (hasarea) {
        pdf.text(738, 502, "Area weapons can hit");
        pdf.text(738, 510, "additional targets in the");
        pdf.text(738, 518, "area with a -2 modifier. Each");
        pdf.text(738, 526, "target can be hit multiple");
        pdf.text(738, 534, "times up to the hits number");
        pdf.text(738, 542, "with one extra hit for each");
        pdf.text(738, 550, "level of victory above minor.");
    } else if (hashits) {
        pdf.text(738, 502, "Each target can be hit multiple");
        pdf.text(738, 510, "times up to the hits number");
        pdf.text(738, 518, "with one extra hit for every");
        pdf.text(738, 525, "level of victory above minor.");
    }

    pdf.setFontSize(8);
    pdf.text(left, 576, chartext);

    // Page 4 - skill table
    pdf.addPage();
    pdf.setFontSize(18);
    pdf.textInBox(character[cid].name, {
        align : "center"
    }, nameleft, namebottom, nameright, nametop);

    var untrained = [];
    untrained.push({
        "skill" : "All untrained skills",
        "level" : 0
    });
    drawSkillTable(pdf, "Skill", untrained, namebottom+8, left, cellheight, cellwidth);
    drawSkillTable(pdf, "Skill", character[cid].sheet.skills, namebottom+54, left, cellheight, cellwidth);

    pdf.setFontSize(8);
    pdf.text(left, 576, chartext);

    return pdf;
};
